/**
 * @file prf_plus.h
 * 
 * @brief Interface for prf_plus.h.
 * 
 */

/*
 * Copyright (C) 2005 Jan Hutter, Martin Willi
 * Hochschule fuer Technik Rapperswil
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#ifndef PRF_PLUS_H_
#define PRF_PLUS_H_


#include <transforms/prfs/prf.h>


typedef struct prf_plus_t prf_plus_t;

/**
 * @brief Implementation of the prf+ function described in IKEv2 draft.
 *
 * This class implements the prf+ algorithm. Internalliy it uses a pseudo random
 * function, which implements the prf_t interface.
 *
 * @see IKEv2 draft 2.13
 * 
 * @ingroup transforms
 */
struct prf_plus_t {
	/**
	 * @brief Get pseudo random bytes.
	 * 
	 * Get the next few bytes of the prf+ output. Space
	 * must be allocated by the caller.
	 * 
	 * @param this			calling prf_plus
	 * @param length		number of bytes to get
	 * @param[out] buffer	pointer where the generated bytes will be written
	 */
	void (*get_bytes) (prf_plus_t *this, size_t length, u_int8_t *buffer);
	
	/**
	 * @brief Allocate pseudo random bytes.
	 * 
	 * Get the next few bytes of the prf+ output. This function
	 * will allocate the required space.
	 * 
	 * @param this			calling prf_plus
	 * @param length		number of bytes to get
	 * @param[out] chunk	chunk which will hold generated bytes
	 */
	void (*allocate_bytes) (prf_plus_t *this, size_t length, chunk_t *chunk);
	
	/**
	 * @brief Destroys a prf_plus_t object.
	 *
	 * @param this 			prf_plus_t object to destroy
	 */
	void (*destroy) (prf_plus_t *this);
};

/**
 * @brief Creates a new prf_plus_t object.
 * 
 * Seed will be cloned. prf will
 * not be cloned, must be destroyed outside after
 * prf_plus usage.
 * 
 * @param prf				prf object to use
 * @param seed				input seed for prf
 * @return					created prf_plus_t
 * 
 * @ingroup transforms
 */
prf_plus_t *prf_plus_create(prf_t *prf, chunk_t seed);

#endif /*PRF_PLUS_H_*/
